# -*- coding: utf-8 -*-
# 2023/2/8
# create by: snower

import os
import copy
import re
import sys
import uuid
import logging.config
import json
import pytz
from syncany.taskers.core import CoreTasker
from syncany.taskers.config import load_config
from syncany.logger import get_logger
from syncany.utils import set_timezone
from .utils import SequenceTypes


VIRTUAL_VIEW_ARGS_RE = re.compile(r"(\#\{\w+?(:.*?)?\})", re.DOTALL | re.M)
CONST_CONFIG_KEYS = ("@verbose", "@timeout", "@limit", "@batch", "@streaming", "@recovery", "@join_batch",
                     "@insert_batch", "@primary_order")


class GlobalConfig(object):
    def __init__(self):
        self.config = copy.deepcopy(CoreTasker.DEFAULT_CONFIG)

    def get_home(self):
        if "SYNCANY_HOME" in os.environ:
            return os.path.abspath(os.environ["SYNCANY_HOME"])
        return os.path.abspath(os.path.join(os.path.expanduser('~'), ".syncany"))

    def get(self):
        return self.config

    def set(self, key, value):
        if not isinstance(key, SequenceTypes):
            key = str(key).split(".")
        if not key:
            return
        if len(key) == 1:
            if value is None:
                self.config.pop(key[0], None)
            else:
                self.config[key[0]] = value
            return
        if key[0] == "databases":
            set_database = None
            for database in self.config["databases"]:
                if database["name"] == key[1]:
                    set_database = database
                    break
            if not set_database:
                set_database = {"name": key[1]}
                self.config["databases"].append(set_database)
            if len(key) <= 2:
                if value is None:
                    self.config["databases"].remove(set_database)
                else:
                    try:
                        set_database.clear()
                        set_database["name"] = key[1]
                        set_database.update(json.loads(value))
                    except:
                        pass
            elif key[2] == "virtual_views":
                if "virtual_views" not in set_database:
                    set_database["virtual_views"] = []
                set_virtual_view = None
                for virtual_view in set_database["virtual_views"]:
                    if virtual_view["name"] == key[3]:
                        set_virtual_view = virtual_view
                        break
                if not set_virtual_view:
                    set_virtual_view = {"name": key[3], "query": "", "args": []}
                    set_database["virtual_views"].append(set_virtual_view)
                if len(key) > 4:
                    if value is None:
                        set_virtual_view.pop(key[4], None)
                    else:
                        try:
                            set_virtual_view[key[4]] = json.loads(value)
                        except:
                            set_virtual_view[key[4]] = value
                else:
                    if value is None:
                        set_database["virtual_views"].remove(set_virtual_view)
                    else:
                        set_virtual_view["query"], set_virtual_view["args"] = self.parse_virtual_view_args(value)
            else:
                if value is None:
                    set_database.pop(key[2], None)
                else:
                    try:
                        set_database[key[2]] = json.loads(value)
                    except:
                        pass
        elif key[0] in ("imports", "sources"):
            if key[0] not in self.config:
                self.config[key[0]] = {}
            if value is None:
                self.config[key[0]].pop(key[1], None)
            else:
                self.config[key[0]][key[1]] = value
        elif key[0] in ("defines", "variables", "options", "caches"):
            if key[0] not in self.config:
                self.config[key[0]] = {}
            if value is None:
                self.config[key[0]].pop(key[1], None)
            else:
                try:
                    self.config[key[0]][key[1]] = json.loads(value)
                except:
                    self.config[key[0]][key[1]] = value

    def parse_virtual_view_args(self, query):
        args = []
        exps = {"eq": "==", "neq": "!=", "gt": ">", "gte": ">=", "lt": "<", "lte": "<=", "in": "in"}
        for arg, default_value in VIRTUAL_VIEW_ARGS_RE.findall(query):
            arg_info = arg[2:-1].split(":")[0].split("__")
            try:
                default_value = json.loads(default_value[1:])
            except:
                default_value = default_value[1:]
            args.append([arg_info[0], exps[arg_info[1]] if len(arg_info) >= 2 else "==", default_value])
            query = query.replace(arg, '%s')
        return query, args

    def load(self, custom_config=None):
        home_config_path, cwd_config_path = self.get_home(), os.path.abspath(os.getcwd())
        if home_config_path not in sys.path:
            sys.path.append(home_config_path)
        for filename in {os.path.join(home_config_path, "config.json"), os.path.join(home_config_path, "config.yaml"),
                         os.path.join(cwd_config_path, "config.json"), os.path.join(cwd_config_path, "config.yaml")}:
            if not os.path.exists(filename):
                continue
            if "extends" not in self.config or not isinstance(self.config["extends"], list):
                self.config["extends"] = []
            if filename not in self.config["extends"]:
                self.config["extends"].append(filename)
        self.load_config()
        if custom_config and isinstance(custom_config, dict):
            self.merge_config(custom_config)
        pypackage_paths = self.config.get("pypackage_paths")
        if pypackage_paths:
            if isinstance(pypackage_paths, str):
                pypackage_paths = [pypackage_paths]
            if isinstance(pypackage_paths, list):
                for pypackage_path in pypackage_paths:
                    if isinstance(pypackage_path, str) and os.path.exists(pypackage_path) \
                            and os.path.isdir(pypackage_path) and pypackage_path not in sys.path:
                        sys.path.append(pypackage_path)

        if "timezone" in self.config:
            timezone = self.config.get("timezone")
            set_timezone(pytz.timezone(timezone))
        if "databases" not in self.config:
            self.config["databases"] = []
        databases = {database["name"]: database for database in self.config["databases"]}
        if "-" not in databases:
            self.config["databases"].append({"name": "-", "driver": "textline"})
        if "--" not in databases:
            self.config["databases"].append({"name": "--", "driver": "memory"})
        elif databases["--"]["driver"] == "redis":
            redis_default_config = {"prefix": "syncany:" + str(uuid.uuid1().int) + ":", "serialize": "pickle",
                                    "ignore_serialize_error": True, "expire_seconds": 900}
            for key, value in redis_default_config.items():
                if key not in databases["--"]:
                    databases["--"][key] = value

        encoding = self.config.get("encoding", None)
        datetime_format = self.config.get("datetime_format", None)
        date_format = self.config.get("date_format", None)
        time_format = self.config.get("time_format", None)
        for database in self.config["databases"]:
            if database["driver"] not in ("textline", "json", "csv"):
                continue
            if encoding:
                database["encoding"] = encoding
            if datetime_format:
                database["datetime_format"] = datetime_format
            if date_format:
                database["date_format"] = date_format
            if time_format:
                database["time_format"] = time_format

        init_execute_files = []
        for init_execute_file in self.config.pop("executes", []):
            if os.path.exists(init_execute_file):
                init_execute_files.append(init_execute_file)
            elif os.path.exists(os.path.join(home_config_path, init_execute_file)):
                init_execute_files.append(os.path.join(home_config_path, init_execute_file))
            else:
                raise FileNotFoundError(init_execute_file)
        return init_execute_files

    def load_config(self, filename=None):
        if filename is None:
            config, self.config = self.config, copy.deepcopy(CoreTasker.DEFAULT_CONFIG)
            for k, v in CoreTasker.DEFAULT_CONFIG.items():
                if k in config and not config[k]:
                    config.pop(k)
        else:
            config = load_config(filename)

        extends = config.pop("extends") if "extends" in config else []
        if isinstance(extends, list):
            for config_filename in extends:
                self.load_config(config_filename)
        else:
            self.load_config(extends)
        self.merge_config(config)

    def load_extensions(self):
        extensions = self.config.get("extensions")
        if not extensions or not isinstance(extensions, list):
            return
        for extension in extensions:
            try:
                __import__(extension, globals(), locals(), [extension.rpartition(".")[-1]])
            except Exception as e:
                get_logger().warning("import extension %s error %s:%s", extension, e.__class__.__name__, str(e))

    def merge_config(self, config):
        for k, v in config.items():
            if k in ("arguments", "imports", "defines", "variables", "sources", "logger", "options"):
                if not isinstance(v, dict) or not isinstance(self.config.get(k, {}), dict):
                    continue

                if k not in self.config:
                    self.config[k] = v
                else:
                    self.config[k].update(v)
            elif k in ("databases", "caches"):
                if not isinstance(v, list) or not isinstance(self.config.get(k, []), list):
                    continue

                if k not in self.config:
                    self.config[k] = v
                else:
                    vs = {c["name"]: c for c in self.config[k]}
                    for d in v:
                        if d["name"] in vs:
                            vs[d["name"]].update(d)
                        else:
                            self.config[k].append(d)
            elif k == "pipelines":
                if self.config[k]:
                    pipelines = copy.copy(self.config[k] if isinstance(self.config[k], list)
                                            and not isinstance(self.config[k][0], str) else [self.config[k]])
                else:
                    pipelines = []
                if v:
                    for pipeline in (v if isinstance(v, list) and not isinstance(v[0], str) else [v]):
                        pipelines.append(pipeline)
                self.config[k] = pipelines
            elif k in ("states", "executes"):
                if k in self.config and self.config[k] and isinstance(self.config[k], list):
                    self.config[k].extend(v if isinstance(v, list) else [v])
                else:
                    self.config[k] = v if isinstance(v, list) else [v]
            else:
                self.config[k] = v

    def config_logging(self, isatty=True):
        logfile = self.config.get("logfile", os.environ.get("SYNCANY_LOGFILE", None))
        if not logfile or logfile == "-":
            logfile = None
        logformat = self.config.get("logformat", "%(asctime)s %(process)d %(levelname)s %(message)s")
        loglevel = self.config.get("loglevel", os.environ.get("SYNCANY_LOGLEVEL", None))
        if "logger" in self.config and isinstance(self.config["logger"], dict):
            logging.config.dictConfig(self.config["logger"])
        else:
            if not logfile and not loglevel and not isatty:
                loglevel = logging.CRITICAL
            else:
                loglevel = {"CRITICAL": logging.CRITICAL, "FATAL": logging.FATAL, "ERROR": logging.ERROR,
                            "WARN": logging.CRITICAL, "WARNING": logging.WARNING, "INFO": logging.INFO,
                            "DEBUG": logging.DEBUG}.get(loglevel or "INFO")
            logging.basicConfig(filename=logfile, level=loglevel, format=logformat if logformat else None,
                                datefmt='%Y-%m-%d %H:%M:%S', filemode='a+')

    def session(self):
        return SessionConfig(self, self)


class SessionConfig(GlobalConfig):
    TASKER_EXTEND_KEYS = ("extends", "databases", "caches", "imports", "sources", "defines", "variables",
                          "options", "states")

    def __init__(self, global_config, parent_config):
        super(SessionConfig, self).__init__()

        self.global_config = global_config
        self.parent_config = parent_config
        self.merged_config = None
        self.tasker_extend_config = None

    def get(self):
        if self.merged_config is None:
            session_config, self.config = self.config, copy.deepcopy(CoreTasker.DEFAULT_CONFIG)
            try:
                self.merge_config(copy.deepcopy(self.parent_config.get()))
                self.merge_config(copy.deepcopy(session_config))
            finally:
                self.merged_config, self.config = self.config, session_config
        return self.merged_config

    def set(self, key, value):
        super(SessionConfig, self).set(key, value)
        self.merged_config = None
        self.tasker_extend_config = None

    def session(self):
        return SessionConfig(self.global_config, self)

    def create_tasker_config(self):
        return copy.deepcopy(CoreTasker.DEFAULT_CONFIG)

    def get_tasker_extend_config(self):
        if self.tasker_extend_config is None:
            merged_config = self.get()
            self.tasker_extend_config = {key: merged_config[key] for key in self.TASKER_EXTEND_KEYS
                                         if key in merged_config}
        return self.tasker_extend_config

    def __bool__(self):
        return bool(self.get())

    def __contains__(self, item):
        return item in self.get()

    def __getitem__(self, item):
        return self.get()[item]

    def items(self):
        return self.get().items()



# -*- coding: utf-8 -*-
# 2023/2/15
# create by: snower

import os
import time
import json
from syncany.logger import get_logger
from ..parser import FileParser


class ExecuteTasker(object):
    def __init__(self, config):
        self.config = config
        self.executor = None

    def start(self, name, executor, session_config, manager, arguments):
        start_time = time.time()
        try:
            if self.config["filename"].endswith(".sql") or self.config["filename"].endswith(".sqlx") or self.config["filename"].endswith(".prql"):
                if self.executor is None:
                    from ..executor import Executor
                    self.executor = Executor(manager, session_config.session(), executor)

                get_logger().info("execute file %s", self.config["filename"])
                try:
                    file_parser = FileParser(self.config["filename"])
                    sqls = file_parser.load()
                    with self.executor as executor:
                        executor.run(self.config["filename"], sqls)
                finally:
                    session_config.merged_config = None
            elif self.config["filename"].endswith(".json"):
                if self.executor is None:
                    from ..executor import Executor
                    self.executor = Executor(manager, session_config.session(), executor)

                from .query import QueryTasker
                with open(self.config["filename"], 'r', encoding=os.environ.get("SYNCANYENCODING", "utf-8")) as fp:
                    tasker = QueryTasker(json.loads(fp.read()))
                with self.executor as executor:
                    arguments = {"@verbose": arguments.get("@verbose", False), "@timeout": arguments.get("@timeout", 0),
                                 "@limit": self.executor.env_variables.get("@limit", 0), "@batch": self.executor.env_variables.get("@batch", 0),
                                 "@streaming": arguments.get("@streaming", False), "@recovery": arguments.get("@recovery", False),
                                 "@join_batch": arguments.get("@join_batch", 10000), "@insert_batch": arguments.get("@insert_batch", 0),
                                 "@use_input": self.executor.env_variables.get("@use_input", None), "@use_output": self.executor.env_variables.get("@use_output", None),
                                 "@use_output_type": self.executor.env_variables.get("@use_output_type", None), "@primary_order": False}
                    name = "execute_" + "".join([c if c.isalpha() or c.isdigit() else '_' for c in self.config["filename"]])
                    executor.runners.extend(tasker.start(name, executor, executor.session_config, executor.manager, arguments))
        finally:
            get_logger().info("execute %s start %.2fms", self.config["filename"], (time.time() - start_time) * 1000)
        return [self]

    def run(self, executor, session_config, manager):
        start_time = time.time()
        try:
            if not self.executor:
                try:
                    os.system(self.config["filename"])
                except Exception as e:
                    get_logger().warning("execute system command error: %s", str(e))
                return
            with self.executor as executor:
                return executor.execute()
        finally:
            get_logger().info("execute %s run %.2fms", self.config["filename"], (time.time() - start_time) * 1000)

    def terminate(self):
        if not self.executor:
            return
        self.executor.terminate()
        self.executor = None



# -*- coding: utf-8 -*-
# 2023/3/2
# create by: snower

import socket
import base64
import binascii
from syncany.calculaters import typing_filter
from ...utils import NumberTypes, ensure_str

@typing_filter(str)
def mysql_bin(x):
    if isinstance(x, NumberTypes):
        return bin(int(x))
    if x is None:
        return None
    if x is True:
        return bin(1)
    if not x:
        return bin(0)
    return bin(x)

@typing_filter(str)
def mysql_hex(x):
    if isinstance(x, NumberTypes):
        return hex(int(x))
    if x is None:
        return None
    if x is True:
        return hex(1)
    if not x:
        return hex(0)
    return binascii.b2a_hex(x.encode("utf-8")).decode("utf-8")

@typing_filter(str)
def mysql_unhex(x):
    if x is None:
        return None
    return binascii.a2b_hex(ensure_str(x))

@typing_filter(str)
def mysql_oct(x):
    if x is None:
        return None
    return oct(x)[2:]

@typing_filter(int)
def mysql_ord(x):
    if x is None:
        return None
    return ord(ensure_str(x))

@typing_filter(str)
def mysql_ascii(s):
    if s is None:
        return None
    return sum([ord(c) for c in ensure_str(s)])

@typing_filter(str)
def mysql_char(*args):
    return "".join([chr(arg) for arg in args])

@typing_filter(int)
def mysql_bit_length(s):
    if s is None:
        return None
    return len(s.encode("utf-8")) * 8

@typing_filter(int)
def mysql_length(s):
    if s is None:
        return None
    return len(s.encode("utf-8"))

@typing_filter(int)
def mysql_char_length(s):
    if s is None:
        return None
    return len(ensure_str(s))

@typing_filter(int)
def mysql_character_length(s):
    if s is None:
        return None
    return len(ensure_str(s))

@typing_filter(str)
def mysql_concat(*args):
    return "".join(ensure_str(arg) for arg in args)

@typing_filter(str)
def mysql_concat_ws(sep, *args):
    return sep.join(ensure_str(arg) for arg in args)

@typing_filter(str)
def mysql_insert(s1, x, l, s2):
    s1, s2 = ensure_str(s1), ensure_str(s2)
    return s1[: x - 1] + s2 + s1[x + l - 1:]

@typing_filter(str)
def mysql_lower(s):
    if s is None:
        return None
    return ensure_str(s).lower()

@typing_filter(str)
def mysql_upper(s):
    if s is None:
        return None
    return ensure_str(s).upper()

@typing_filter(str)
def mysql_ucase(s):
    if s is None:
        return None
    return ensure_str(s).upper()

@typing_filter(str)
def mysql_left(s, x):
    if s is None:
        return None
    return ensure_str(s)[:x]

@typing_filter(str)
def mysql_right(s, x):
    if s is None:
        return None
    return ensure_str(s)[-x:]

@typing_filter(str)
def mysql_trim(s):
    if s is None:
        return None
    return ensure_str(s).strip()

@typing_filter(str)
def mysql_elt(n, *args):
    return args[n - 1] if n < len(args) else None

@typing_filter(int)
def mysql_field(s, *args):
    if s is None:
        return None
    try:
        return args.index(s) + 1
    except ValueError:
        return 0

@typing_filter(int)
def mysql_find_in_set(s, ss):
    if s is None:
        return None
    try:
        return ensure_str(ss).split(",").index(s) + 1
    except ValueError:
        return 0

@typing_filter(str)
def mysql_replace(s, s1, s2):
    if s is None:
        return None
    return ensure_str(s).replace(s1, s2)

@typing_filter(str)
def mysql_substring(s, n, l=None):
    if s is None:
        return None
    s = ensure_str(s)
    if isinstance(l, int):
        return s[n - 1: n + l - 1]
    return s[n - 1:]

@typing_filter(str)
def mysql_substr(s, n, l=None):
    if s is None:
        return None
    s = ensure_str(s)
    if isinstance(l, int):
        return s[n - 1: n + l - 1]
    return s[n - 1:]

@typing_filter(str)
def mysql_substring_index(s, d, c):
    if s is None:
        return None
    s, d = ensure_str(s), ensure_str(d)
    if c < 0:
        return d.join(s.split(d)[-c:])
    return d.join(s.split(d)[:c])

@typing_filter(str)
def mysql_repeat(s, c):
    if s is None:
        return None
    return ensure_str(s) * c

@typing_filter(str)
def mysql_reverse(s):
    if s is None:
        return None
    return ensure_str(s)[::-1]

@typing_filter(int)
def mysql_strcmp(s1, s2):
    if s1 is None or s2 is None:
        return None
    s1, s2 = ensure_str(s1), ensure_str(s2)
    return 0 if s1 == s2 else (-1 if s1 < s2 else 1)

@typing_filter(int)
def mysql_startswith(s1, s2):
    if s1 is None or s2 is None:
        return None
    s1, s2 = ensure_str(s1), ensure_str(s2)
    return 1 if s1.startswith(s2) else 0

@typing_filter(int)
def mysql_endswith(s1, s2):
    if s1 is None or s2 is None:
        return None
    s1, s2 = ensure_str(s1), ensure_str(s2)
    return 1 if s1.endswith(s2) else 0

@typing_filter(int)
def mysql_contains(s1, s2):
    if s1 is None or s2 is None:
        return None
    try:
        return 1 if s2 in s1 else 0
    except:
        s1, s2 = ensure_str(s1), ensure_str(s2)
        return 1 if s2 in s1 else 0

@typing_filter(int)
def mysql_crc32(s):
    if s is None:
        return None
    import zlib
    return zlib.crc32(s)

@typing_filter(str)
def mysql_from_base64(s):
    if s is None:
        return None
    return base64.b64decode(s.encode("utf-8")).decode("utf-8")

@typing_filter(str)
def mysql_to_base64(s):
    if s is None:
        return None
    return base64.b64encode(s.encode("utf-8")).decode("utf-8")

@typing_filter(str)
def mysql_inet4_aton(s):
    if s is None:
        return None
    try:
        return binascii.b2a_hex(socket.inet_aton(s)).decode("utf-8")
    except:
        return None

@typing_filter(str)
def mysql_inet4_ntoa(b):
    if b is None:
        return None
    try:
        return socket.inet_ntoa(b)
    except:
        return None

@typing_filter(int)
def mysql_is_ipv4(s):
    if s is None:
        return None
    try:
        socket.inet_aton(s)
        return 1
    except:
        return 0

@typing_filter(str)
def mysql_inet6_aton(s):
    if s is None:
        return None
    try:
        return binascii.b2a_hex(socket.inet_pton(socket.AF_INET6, s)).decode("utf-8")
    except:
        return None

@typing_filter(str)
def mysql_inet6_ntoa(b):
    if b is None:
        return None
    try:
        return socket.inet_ntop(socket.AF_INET6, b)
    except:
        return None

@typing_filter(int)
def mysql_is_ipv6(s):
    if s is None:
        return None
    try:
        socket.inet_pton(socket.AF_INET6, s)
        return 1
    except:
        return 0


funcs = {key[6:]: value for key, value in globals().items() if key.startswith("mysql_")}



# -*- coding: utf-8 -*-
# 2023/4/24
# create by: snower

from .example import ExampleTestCase


class AggregateExampleTestCase(ExampleTestCase):
    example_name = "aggregate"

    def test_aggregate(self):
        self.execute("aggregate.sql")

        self.assert_data(5, [{'cnt': 6, 'total_amount': 43.2, 'avg_amount': 7.2, 'min_amount': 3, 'max_amount': 9.6}],
                         "data error")

        self.assert_data(7, [
            {'uid': 2, 'cnt': 4, 'total_amount': 27.6, 'avg_amount': 6.9, 'min_amount': 3, 'max_amount': 9.6},
            {'uid': 1, 'cnt': 2, 'total_amount': 15.6, 'avg_amount': 7.8, 'min_amount': 7.6, 'max_amount': 8}],
                         "data error")

        self.assert_data(9, [{'uid': 2, 'order_id': 1, 'amount': 9.6}, {'uid': 1, 'order_id': 2, 'amount': 7.6},
                             {'uid': 2, 'order_id': 3, 'amount': 3}, {'uid': 1, 'order_id': 4, 'amount': 8},
                             {'uid': 2, 'order_id': 5, 'amount': 8}, {'uid': 2, 'order_id': 6, 'amount': 7}],
                         "data error")

        self.assert_data(11, [{'ucnt': 2}], "data error")

        self.assert_data(13, [{'uid': 2, 'gcnt': 2}, {'uid': 1, 'gcnt': 1}], "data error")

        self.assert_data(15, [{'uid': 1, 'gcnt': 1}], "data error")

        self.assert_data(17, [{'uid': 2, 'gcnt': 2}], "data error")

        self.assert_data(19, [{'uid': 2, 'avg_amount': 0.069}, {'uid': 1, 'avg_amount': 0.078}], "data error")

        self.assert_data(21, [{'name': '李四', 'goods_name': '青菜', 'cnt': 6}], "data error")

        self.assert_data(24,
                         [{'name': '李四', 'goods_name': '青菜', 'cnt': 2},
                          {'name': '王五', 'goods_name': '青菜', 'cnt': 2},
                          {'name': '李四', 'goods_name': '白菜', 'cnt': 2}], "data error")

        self.assert_data(27, [{'name': '李四', 'goods_name': '青菜'}, {'name': '王五', 'goods_name': '青菜'},
                              {'name': '李四', 'goods_name': '白菜'}], "data error")

        self.assert_data(30,
                         [{'name': '李四', 'goods_name': '青菜', 'cnt': 2},
                          {'name': '李四', 'goods_name': '白菜', 'cnt': 1}],
                         "data error")

        self.assert_data(33, [{'uid': 2, 'cgoods_ids': '1,2,1,2', 'agoods_ids': [1, 2, 1, 2], 'uagoods_ids': [1, 2]},
                              {'uid': 1, 'cgoods_ids': '1,1', 'agoods_ids': [1, 1], 'uagoods_ids': [1]}], "data error")

        self.assert_data(35, [{'uid': 2, 'goods_id_and': 0, 'goods_id_or': 3, 'goods_id_xor': 0},
                              {'uid': 1, 'goods_id_and': 1, 'goods_id_or': 1, 'goods_id_xor': 0}], "data error")

        self.assert_data(38, [{'uids': {1, 2}, 'order_ids': '1,2,3,4,5,6'}], "data error")

        self.assert_data(40, [{'uid': 2, 'goods_ids': {1, 2}, 'order_ids': '1,3,5,6'},
                              {'uid': 1, 'goods_ids': {1}, 'order_ids': '2,4'}], "data error")

        self.assert_data(42, [{'uid': 2, 'avg_amount': 0.07, 'percent': 0.0275},
                              {'uid': 1, 'avg_amount': 0.03, 'percent': 0.025}], "data error")

        self.assert_data(44, [
            {'name': '李四', 'goods_name': '青菜', 'names': {'李四', '王五'},
             'goods_namees': '青菜,青菜,白菜,青菜,青菜,白菜'}],
                         "data error")

        self.assert_data(48, [{'name': '李四', 'goods_name': '青菜', 'names': {'李四'}, 'goods_namees': '青菜,青菜'},
                              {'name': '王五', 'goods_name': '青菜', 'names': {'王五'}, 'goods_namees': '青菜,青菜'},
                              {'name': '李四', 'goods_name': '白菜', 'names': {'李四'}, 'goods_namees': '白菜,白菜'}],
                         "data error")

        self.assert_data(52, [{'uid': 2, 'agoods_ids': '[1, 2, 1, 2]', 'ogoods_ids': '{"1": 5, "2": 6}'},
                              {'uid': 1, 'agoods_ids': '[1, 1]', 'ogoods_ids': '{"1": 4}'}], "data error")

        self.assert_data(54, [{'uid': 2, 'goods_id': 1}, {'uid': 1, 'goods_id': 1}], "data error")

    def test_aggregate_batch(self):
        self.execute("aggregate_batch.sql")

        self.assert_data(5, [{'cnt': 6, 'total_amount': 43.2, 'avg_amount': 7.2, 'min_amount': 3, 'max_amount': 9.6}],
                         "data error")

        self.assert_data(7, [
            {'uid': 2, 'cnt': 4, 'total_amount': 27.6, 'avg_amount': 6.9, 'min_amount': 3, 'max_amount': 9.6},
            {'uid': 1, 'cnt': 2, 'total_amount': 15.6, 'avg_amount': 7.8, 'min_amount': 7.6, 'max_amount': 8}],
                         "data error")

        self.assert_data(9, [{'uid': 2, 'order_id': 1, 'amount': 9.6}, {'uid': 1, 'order_id': 2, 'amount': 7.6},
                             {'uid': 2, 'order_id': 3, 'amount': 3}, {'uid': 1, 'order_id': 4, 'amount': 8},
                             {'uid': 2, 'order_id': 5, 'amount': 8}, {'uid': 2, 'order_id': 6, 'amount': 7}],
                         "data error")

        self.assert_data(11, [{'ucnt': 2}], "data error")

        self.assert_data(13, [{'uid': 2, 'gcnt': 2}, {'uid': 1, 'gcnt': 1}], "data error")

        self.assert_data(15, [{'uid': 1, 'gcnt': 1}], "data error")

        self.assert_data(17, [{'uid': 2, 'gcnt': 2}], "data error")

        self.assert_data(19, [{'uid': 2, 'avg_amount': 0.069}, {'uid': 1, 'avg_amount': 0.078}], "data error")

        self.assert_data(21, [{'name': '李四', 'goods_name': '青菜', 'cnt': 6}], "data error")

        self.assert_data(24,
                         [{'name': '李四', 'goods_name': '青菜', 'cnt': 2},
                          {'name': '王五', 'goods_name': '青菜', 'cnt': 2},
                          {'name': '李四', 'goods_name': '白菜', 'cnt': 2}], "data error")

        self.assert_data(27, [{'name': '李四', 'goods_name': '青菜'}, {'name': '王五', 'goods_name': '青菜'},
                              {'name': '李四', 'goods_name': '白菜'}], "data error")

        self.assert_data(30,
                         [{'name': '李四', 'goods_name': '青菜', 'cnt': 2},
                          {'name': '李四', 'goods_name': '白菜', 'cnt': 1}],
                         "data error")

        self.assert_data(33, [{'uid': 2, 'cgoods_ids': '1,2,1,2', 'agoods_ids': [1, 2, 1, 2], 'uagoods_ids': [1, 2]},
                              {'uid': 1, 'cgoods_ids': '1,1', 'agoods_ids': [1, 1], 'uagoods_ids': [1]}], "data error")

        self.assert_data(35, [{'uid': 2, 'goods_id_and': 0, 'goods_id_or': 3, 'goods_id_xor': 0},
                              {'uid': 1, 'goods_id_and': 1, 'goods_id_or': 1, 'goods_id_xor': 0}], "data error")

        self.assert_data(38, [{'uids': {1, 2}, 'order_ids': '1,2,3,4,5,6'}], "data error")

        self.assert_data(40, [{'uid': 2, 'goods_ids': {1, 2}, 'order_ids': '1,3,5,6'},
                              {'uid': 1, 'goods_ids': {1}, 'order_ids': '2,4'}], "data error")

        self.assert_data(42, [{'uid': 2, 'avg_amount': 0.07}, {'uid': 1, 'avg_amount': 0.03}], "data error")

        self.assert_data(44, [
            {'name': '李四', 'goods_name': '青菜', 'names': {'王五', '李四'},
             'goods_namees': '青菜,青菜,白菜,青菜,青菜,白菜'}],
                         "data error")

        self.assert_data(48, [{'name': '李四', 'goods_name': '青菜', 'names': {'李四'}, 'goods_namees': '青菜,青菜'},
                              {'name': '王五', 'goods_name': '青菜', 'names': {'王五'}, 'goods_namees': '青菜,青菜'},
                              {'name': '李四', 'goods_name': '白菜', 'names': {'李四'}, 'goods_namees': '白菜,白菜'}],
                         "data error")

    def test_window_aggregate(self):
        self.execute("window_aggregate.sql")

        self.assert_data(1, [{'cnt': 2,
                              'goods_name': '青菜',
                              'name': '李四',
                              'order_id': 1,
                              'order_time': '2024-10-01 10:09:10',
                              'total_amount': 24.3,
                              'unorder_time': '2024-10-06 11:09:10'},
                             {'cnt': 1,
                              'goods_name': '青菜',
                              'name': '王五',
                              'order_id': 2,
                              'order_time': '2024-10-01 12:09:10',
                              'total_amount': 7.6,
                              'unorder_time': None},
                             {'cnt': 2,
                              'goods_name': '白菜',
                              'name': '李四',
                              'order_id': 3,
                              'order_time': '2024-10-02 09:09:10',
                              'total_amount': 34.4,
                              'unorder_time': '2024-10-08 09:09:10'},
                             {'cnt': 1,
                              'goods_name': '青菜',
                              'name': '王五',
                              'order_id': 4,
                              'order_time': '2024-10-01 10:09:10',
                              'total_amount': 8,
                              'unorder_time': None},
                             {'cnt': 2,
                              'goods_name': '青菜',
                              'name': '李四',
                              'order_id': 5,
                              'order_time': '2024-10-01 10:09:10',
                              'total_amount': 100.80000000000001,
                              'unorder_time': '2024-10-06 15:09:10'},
                             {'cnt': 1,
                              'goods_name': '白菜',
                              'name': '李四',
                              'order_id': 6,
                              'order_time': '2024-10-01 10:09:10',
                              'total_amount': 7,
                              'unorder_time': None}],
                         "data error")

        self.assert_data(14, [{'first_create_time': '2024-10-01 10:09:10',
                               'history_type': 1,
                               'id': 1,
                               'next_create_time': '2024-10-02 14:09:10',
                               'next_history_type': 0,
                               'order_id': 1},
                              {'first_create_time': '2024-10-01 10:09:10',
                               'history_type': 0,
                               'id': 2,
                               'next_create_time': '2024-10-06 11:09:10',
                               'next_history_type': 1,
                               'order_id': 1},
                              {'first_create_time': '2024-10-06 11:09:10',
                               'history_type': 1,
                               'id': 4,
                               'next_create_time': None,
                               'next_history_type': None,
                               'order_id': 1}],
                         "data error")

        self.assert_data(22, [{'cnt': 3,
                               'history_type': 1,
                               'id': 1,
                               'next_create_time': None,
                               'next_history_type': None,
                               'order_id': 1,
                               'total_amount': 33.900000000000006},
                              {'cnt': 1,
                               'history_type': 1,
                               'id': 3,
                               'next_create_time': None,
                               'next_history_type': None,
                               'order_id': 2,
                               'total_amount': 7.6},
                              {'cnt': 5,
                               'history_type': 1,
                               'id': 5,
                               'next_create_time': None,
                               'next_history_type': None,
                               'order_id': 3,
                               'total_amount': 44.7},
                              {'cnt': 1,
                               'history_type': 1,
                               'id': 8,
                               'next_create_time': None,
                               'next_history_type': None,
                               'order_id': 4,
                               'total_amount': 8},
                              {'cnt': 3,
                               'history_type': 1,
                               'id': 11,
                               'next_create_time': None,
                               'next_history_type': None,
                               'order_id': 5,
                               'total_amount': 108.80000000000001},
                              {'cnt': 1,
                               'history_type': 1,
                               'id': 12,
                               'next_create_time': None,
                               'next_history_type': None,
                               'order_id': 6,
                               'total_amount': 7}],
                         "data error")

        self.assert_data(27, [{'cnt': 2,
                              'goods_name': '青菜',
                              'name': '李四',
                              'order_id': 1,
                              'order_time': '2024-10-01 10:09:10',
                              'total_amount': 24.3,
                              'unorder_time': '2024-10-06 11:09:10'},
                             {'cnt': 2,
                              'goods_name': '白菜',
                              'name': '李四',
                              'order_id': 3,
                              'order_time': '2024-10-02 09:09:10',
                              'total_amount': 34.4,
                              'unorder_time': '2024-10-08 09:09:10'},
                             {'cnt': 2,
                              'goods_name': '青菜',
                              'name': '李四',
                              'order_id': 5,
                              'order_time': '2024-10-01 10:09:10',
                              'total_amount': 100.80000000000001,
                              'unorder_time': '2024-10-06 15:09:10'}],
                         "data error")



# -*- coding: utf-8 -*-
# 2023/3/14
# create by: snower

import json
from syncany.calculaters import typing_filter

ArrayStringTypes = (list, str)
ArrayObjectTypes = (list, dict)


def parse_json_path(json_path):
    if not json_path:
        return []
    try:
        dot_index = json_path.index(".")
    except ValueError:
        dot_index = -1

    json_path_keys = []
    try:
        start_index = json_path.index("[")
        if 0 < dot_index < start_index:
            json_path_keys.append(json_path[:dot_index])
            next_json_path = json_path[dot_index + 1:].strip(".")
            if next_json_path:
                json_path_keys.extend(parse_json_path(next_json_path))
            return json_path_keys

        end_index = json_path.index("]")
        if start_index > 0:
            json_path_keys.append(json_path[:start_index])
        if end_index - start_index > 0:
            json_path_key = json_path[start_index + 1: end_index]
            if json_path_key.isdigit():
                json_path_key = int(json_path_key)
            elif json_path_key and json_path_key[0] == '"' and json_path_key[-1] == '"':
                json_path_key = json_path_key[1:-1]
            json_path_keys.append(json_path_key)

        next_json_path = json_path[end_index + 1:].strip(".")
        if next_json_path:
            json_path_keys.extend(parse_json_path(next_json_path))
        return json_path_keys
    except ValueError:
        if dot_index > 0:
            json_path_keys.append(json_path[:dot_index])
            next_json_path = json_path[dot_index + 1:].strip(".")
            if next_json_path:
                json_path_keys.extend(parse_json_path(next_json_path))
        else:
            json_path_keys.append(json_path)
        return json_path_keys


def get_json_path_value(json_data, json_path):
    if json_data is None:
        raise ValueError('value is None')
    if json_path and json_path[:2] == "$.":
        json_path = json_path[2:]
    elif json_path and json_path[0] == '$':
        json_path = json_path[1:]

    def get_value(json_data, json_path_keys):
        for i in range(len(json_path_keys)):
            json_path_key = json_path_keys[i]
            if isinstance(json_path_key, int):
                if isinstance(json_data, ArrayStringTypes):
                    if json_path_key >= len(json_data):
                        return None
                    json_data = json_data[json_path_key]
                    continue
                return None
            if json_path_key == "*":
                if i + 1 >= len(json_path_keys):
                    return json_data
                if not isinstance(json_data, list):
                    json_data = [json_data]
                result = []
                for d in json_data:
                    v = get_value(d, json_path_keys[i + 1:])
                    if v is None:
                        continue
                    result.append(v)
                return result
            if not isinstance(json_data, dict):
                return None
            if json_path_key not in json_data:
                return None
            json_data = json_data[json_path_key]
        return json_data
    return get_value(json_data, parse_json_path(json_path))


def set_json_path_value(json_data, json_path, value):
    if json_data is None:
        raise ValueError('value is None')
    if json_path and json_path[:2] == "$.":
        json_path = json_path[2:]
    elif json_path and json_path[0] == '$':
        json_path = json_path[1:]

    def set_value(json_data, json_path_keys):
        for i in range(len(json_path_keys)):
            json_path_key = json_path_keys[i]
            if isinstance(json_path_key, int):
                if isinstance(json_data, list):
                    if i + 1 >= len(json_path_keys):
                        if json_path_key >= len(json_data):
                            json_data.append(value)
                            return json_data
                        json_data[json_path_key] = value
                        return json_data
                    json_data = [] if json_path_key >= len(json_data) else json_data[json_path_key]
                elif isinstance(json_data, str):
                    if i + 1 >= len(json_path_keys):
                        if json_path_key >= len(json_data):
                            return [json_data, value]
                        return json_data[:json_path_key] + str(value) + json_data[json_path_key + 1:]
                    json_data = [] if json_path_key >= len(json_data) else json_data[json_path_key]
                else:
                    json_data = []
                continue
            if not isinstance(json_data, dict):
                return json_data
            if json_path_key not in json_data:
                if i + 1 >= len(json_path_keys):
                    json_data[json_path_key] = value
                return json_data
            if i + 1 >= len(json_path_keys):
                json_data[json_path_key] = value
                return json_data
            json_data = json_data[json_path_key]
        return json_data
    return set_value(json_data, parse_json_path(json_path))


def remove_json_path_value(json_data, json_path):
    if json_data is None:
        raise ValueError('value is None')
    if json_path and json_path[:2] == "$.":
        json_path = json_path[2:]
    elif json_path and json_path[0] == '$':
        json_path = json_path[1:]

    def remove_value(json_data, json_path_keys):
        for i in range(len(json_path_keys)):
            json_path_key = json_path_keys[i]
            if isinstance(json_path_key, int):
                if isinstance(json_data, list):
                    if i + 1 >= len(json_path_keys):
                        if json_path_key >= len(json_data):
                            return json_data
                        del json_data[json_path_key]
                        return json_data
                    if json_path_key < len(json_data):
                        del json_data[json_path_key]
                continue
            if not isinstance(json_data, dict):
                return json_data
            if json_path_key not in json_data:
                return json_data
            if i + 1 >= len(json_path_keys):
                json_data.pop(json_path_key)
                return json_data
            json_data = json_data[json_path_key]
        return json_data
    return remove_value(json_data, parse_json_path(json_path))

def mysql_json_encode(value):
    return json.dumps(value, default=str, ensure_ascii=False)

def mysql_json_decode(content):
    if not isinstance(content, str):
        return None
    return json.loads(content)

def mysql_json_array(*values):
    return list(values)

def mysql_json_object(*values):
    obj = {}
    for i in range(int(len(values) / 2)):
        obj[values[i * 2]] = values[i * 2 + 1]
    return obj

def json_array_append(json_doc, *path_vals):
    if isinstance(json_doc, str):
        try:
            json_doc = json.loads(json_doc)
        except:
            pass
    for i in range(int(len(path_vals) / 2)):
        json_value = get_json_path_value(json_doc, path_vals[i * 2])
        if isinstance(json_value, list):
            json_value.append(path_vals[i * 2 + 1])
    return json_doc

@typing_filter(int)
def mysql_json_contains(target, candidate, path):
    if isinstance(target, str):
        try:
            target = json.loads(target)
        except:
            pass
    if isinstance(candidate, str):
        try:
            candidate = json.loads(candidate)
        except:
            pass
    target_value = get_json_path_value(target, path)
    return 1 if target_value == candidate else 0


@typing_filter(int)
def mysql_json_contains_path(json_doc, one_or_all, *paths):
    if isinstance(json_doc, str):
        try:
            json_doc = json.loads(json_doc)
        except:
            pass
    if not one_or_all:
        one_or_all = "one"
    for path in paths:
        json_value = get_json_path_value(json_doc, path)
        if one_or_all.lower() == "one":
            if json_value is not None:
                return 1
        elif json_value is None:
            return 0
    return 0 if one_or_all.lower() == "one" else 1


def mysql_json_extract(json_doc, *paths):
    if isinstance(json_doc, str):
        try:
            json_doc = json.loads(json_doc)
        except:
            pass
    results = []
    for path in paths:
        results.append(get_json_path_value(json_doc, path))
    return results[0] if len(paths) == 1 else results

def mysql_json_set(json_doc, *path_vals):
    if isinstance(json_doc, str):
        try:
            json_doc = json.loads(json_doc)
        except:
            pass
    for i in range(int(len(path_vals) / 2)):
        if not isinstance(json_doc, ArrayObjectTypes):
            json_doc = set_json_path_value(json_doc, path_vals[i * 2], path_vals[i * 2 + 1])
        else:
            set_json_path_value(json_doc, path_vals[i * 2], path_vals[i * 2 + 1])
    return json_doc

def mysql_json_remove(json_doc, *paths):
    if isinstance(json_doc, str):
        try:
            json_doc = json.loads(json_doc)
        except:
            pass
    for path in paths:
        remove_json_path_value(json_doc, path)
    return json_doc

@typing_filter(int)
def mysql_json_depth(json_doc):
    if isinstance(json_doc, str):
        try:
            json_doc = json.loads(json_doc)
        except:
            pass
    def get_depth(json_doc):
        if not json_doc:
            return 1
        if isinstance(json_doc, list):
            return max((get_depth(json_value) for json_value in json_doc)) + 1
        if isinstance(json_doc, dict):
            return max((get_depth(json_value) for json_value in json_doc.values())) + 1
        return 1
    return get_depth(json_doc)


@typing_filter(list)
def mysql_json_keys(json_doc, path=None):
    if isinstance(json_doc, str):
        try:
            json_doc = json.loads(json_doc)
        except:
            pass
    if path:
        json_doc = get_json_path_value(json_doc, path)
    if not isinstance(json_doc, dict):
        return []
    return list(json_doc.keys())


@typing_filter(int)
def mysql_json_length(json_doc, path=None):
    if isinstance(json_doc, str):
        try:
            json_doc = json.loads(json_doc)
        except:
            pass
    if path:
        json_doc = get_json_path_value(json_doc, path)
    if not isinstance(json_doc, ArrayObjectTypes):
        return 0
    return len(json_doc)


@typing_filter(int)
def mysql_json_valid(val):
    if isinstance(val, str):
        try:
            json.loads(val)
            return 1
        except:
            return 0
    return 1


funcs = {key[6:]: value for key, value in globals().items() if key.startswith("mysql_")}



# -*- coding: utf-8 -*-
# 23/02/07
# create by: snower

import sys
import os
from setuptools import find_packages, setup

version = "0.1.28"

if os.path.exists("README.md"):
    if sys.version_info[0] >= 3:
        try:
            with open("README.md", encoding="utf-8") as fp:
                long_description = fp.read()
        except Exception as e:
            print("Waring: " + str(e))
            long_description = 'https://github.com/snower/syncany-sql'
    else:
        try:
            with open("README.md") as fp:
                long_description = fp.read()
        except Exception as e:
            print("Waring: " + str(e))
            long_description = 'https://github.com/snower/syncany-sql'
else:
    long_description = 'https://github.com/snower/syncany-sql'

setup(
    name='syncanysql',
    version=version,
    url='https://github.com/snower/syncany-sql',
    author='snower',
    author_email='sujian199@gmail.com',
    license='MIT',
    packages=find_packages(exclude=['*tests*']),
    zip_safe=False,
    install_requires=[
        "pyyaml>=6.0",
        "sqlglot>=11.5.5,<12",
        "syncany>=0.2.29",
        'Pygments>=2.14.0',
        'prompt-toolkit>=3.0.36',
        "rich>=9.11.1",
    ],
    extras_require={
        "pymongo": ['pymongo>=3.6.1'],
        "pymysql": ['PyMySQL>=0.8.1'],
        "openpyxl": ["openpyxl>=2.5.0"],
        "postgresql": ["psycopg2>=2.8.6"],
        "elasticsearch": ["elasticsearch>=6.3.1"],
        "influxdb": ["influxdb>=5.3.1"],
        "clickhouse": ["clickhouse_driver>=0.1.5"],
        "redis": ["redis>=3.5.3"],
        "requests": ["requests>=2.22.0"],
        "pymssql": ['pymssql>=2.2.7'],
        "prql-python": ["prql-python>=0.11.1"],
    },
    package_data={
        '': ['README.md'],
    },
    entry_points={
        'console_scripts': [
            'syncany-sql = syncanysql.main:main',
        ],
    },
    description='Simple and easy-to-use sql execution engine',
    long_description=long_description,
    long_description_content_type='text/markdown'
)



# -*- coding: utf-8 -*-
# 2023/2/8
# create by: snower

import os
import copy
import time
import traceback
import uuid
from collections import defaultdict
from syncany.logger import get_logger
from syncany.taskers.core import CoreTasker
from syncany.hook import Hooker
from syncany.taskers.tasker import _thread_local
from syncany.main import TaskerYieldNext, warp_database_logging


DEAULT_AGGREGATE = {"key": None, "schema": {}, "window_schema": {}, "having_columns": set([]),
                    "distinct_keys": [], "distinct_aggregates": set([])}


class ReduceHooker(Hooker):
    def __init__(self, executor, session_config, manager, arguments, tasker, config, batch, aggregate):
        self.executor = executor
        self.session_config = session_config
        self.manager = manager
        self.arguments = arguments
        self.tasker = tasker
        self.config = config
        self.batch = batch
        self.aggregate = aggregate
        self.count = 0
        self.batch_count = 0

    def outputed(self, tasker, datas):
        if not datas:
            return
        if self.batch > 0 and self.count > 0 and self.count + len(datas) >= self.batch:
            self.count, self.batch_count = 0, 1
            self.tasker.run_reduce(self.executor, self.session_config, self.manager, self.arguments, False)
        else:
            self.count += len(datas)
            self.batch_count += 1
            limit = tasker.arguments["@limit"] if "@limit" in tasker.arguments and tasker.arguments["@limit"] > 0 else 0
            if 0 < limit <= tasker.status["store_count"]:
                self.count, self.batch_count = 0, 1
                self.tasker.run_reduce(self.executor, self.session_config, self.manager, self.arguments, False)

    def finaled(self, tasker, e=None):
        if e is not None or self.batch_count <= 1:
            return
        self.tasker.run_reduce(self.executor, self.session_config, self.manager, self.arguments, False)


class QueryTaskerTemporaryMemoryManager(object):
    def __init__(self):
        self.collections = defaultdict(int)

    def add_collection(self, name):
        if name not in self.collections:
            self.collections[name] = 0

    def use_collection(self, name):
        self.collections[name] += 1

    def unuse_collection(self, name):
        if name not in self.collections:
            return True
        self.collections[name] -= 1
        return self.collections[name] <= 0

    def free_collection(self, name):
        if name not in self.collections:
            return True
        if self.collections[name] > 0:
            return False
        self.collections.pop(name, None)
        return True

    def get_names(self):
        return list(self.collections.keys())


class QueryTasker(object):
    ENV_ARGUMENTS = {key.lower(): value for key, value in os.environ.items()}

    def __init__(self, config, sql_expression=None, temporary_memory_manager=None, is_inner_subquery=False):
        self.name = config.get("name", "")
        self.config = config
        self.sql_expression = sql_expression
        self.reduce_config = None
        self.tasker = None
        self.dependency_taskers = []
        self.arguments = None
        self.tasker_generator = None
        self.updaters = []
        self.temporary_memory_manager = temporary_memory_manager or QueryTaskerTemporaryMemoryManager()
        self.is_inner_subquery = is_inner_subquery
        self.run_start_time = 0
        self.tasker_index = 0

    def start(self, name, executor, session_config, manager, arguments):
        dependency_taskers, aggregate = [], self.config.pop("aggregate", None)
        if aggregate and aggregate.get("distinct_keys"):
            self.config, distinct_config = self.compile_distinct_config(aggregate), self.config
            distinct_config["name"] = distinct_config["name"] + "#select@distinct"
            distinct_tasker = QueryTasker(distinct_config, temporary_memory_manager=self.temporary_memory_manager, is_inner_subquery=True)
            distinct_tasker.start(name, executor, session_config, manager, copy.deepcopy(arguments))
            dependency_taskers.append(distinct_tasker)
            arguments["@limit"] = 0

        for dependency_config in self.config.get("dependencys", []):
            kn, knl = (dependency_config["name"] + "@"), len(dependency_config["name"] + "@")
            dependency_arguments = {}
            for key, value in arguments.items():
                if key[:knl] != kn:
                    continue
                dependency_arguments[key[knl:]] = value
            dependency_tasker = QueryTasker(dependency_config, temporary_memory_manager=self.temporary_memory_manager, is_inner_subquery=True)
            dependency_tasker.start(name, executor, session_config, manager, dependency_arguments)
            dependency_taskers.append(dependency_tasker)

        limit, batch = int(arguments.get("@limit", 0)), int(arguments.get("@batch", 0))
        require_reduce = False
        if aggregate and (aggregate.get("schema") or aggregate["window_schema"]):
            if batch > 0:
                require_reduce = True
            elif [aggregate_column["final_value"] for aggregate_column in aggregate["schema"].values()
                  if aggregate_column["final_value"]]:
                require_reduce = True
            elif [aggregate_column["final_value"] for aggregate_column in aggregate["window_schema"].values()
                  if aggregate_column["final_value"]]:
                require_reduce = True
        if require_reduce and isinstance(self.config["schema"], dict) and not arguments.get("@streaming"):
            if not aggregate:
                aggregate = copy.deepcopy(DEAULT_AGGREGATE)
            self.compile_reduce_config(aggregate)
        elif 0 < limit < batch:
            arguments["@batch"] = limit

        if self.tasker_index <= 0:
            self.tasker_index = executor.distribute_tasker_index()
        tasker = CoreTasker(self.config, manager)
        if "#" not in tasker.config["name"]:
            tasker.config["name"] = tasker.config["name"] + "#select"
        if self.reduce_config and "_aggregate_key_" in self.reduce_config["schema"]:
            tasker.add_hooker(ReduceHooker(executor, session_config, manager, arguments,
                                           self, copy.deepcopy(self.config), batch, aggregate))
        arguments = self.compile_tasker(arguments, tasker)
        self.tasker, self.dependency_taskers, self.arguments = tasker, dependency_taskers, arguments
        return [self]

    def run(self, executor, session_config, manager):
        try:
            self.execute_updater(executor, session_config, manager)
            if not self.tasker_generator:
                self.tasker_generator = self.run_tasker(executor, session_config, manager, self.tasker, self.dependency_taskers)
                self.run_start_time = time.time()
            else:
                _thread_local.current_tasker = self.tasker
            while True:
                try:
                    value = self.tasker_generator.send(None)
                    if isinstance(value, TaskerYieldNext):
                        executor.add_runner(self)
                        return
                except StopIteration:
                    break
                except Exception as e:
                    if self.sql_expression:
                        get_logger().info("execute SQL %s finish with Exception %s:%s %.2fms", self.name, e.__class__.__name__, e,
                                          (time.time() - self.run_start_time) * 1000)
                    raise e
            if self.reduce_config:
                try:
                    self.run_reduce(executor, session_config, manager, self.arguments, True)
                except Exception as e:
                    if self.sql_expression:
                        get_logger().info("execute SQL %s finish with Exception %s:%s %.2fms", self.name, e.__class__.__name__, e,
                                          (time.time() - self.run_start_time) * 1000)
                    raise e
                if self.sql_expression:
                    get_logger().info("execute SQL %s finish %.2fms", self.name, (time.time() - self.run_start_time) * 1000)
        finally:
            names = self.get_temporary_memory_collections()
            if names:
                executor.clear_temporary_memory_collection(names)

    def run_yield(self, executor, session_config, manager, tasker, dependency_taskers):
        tasker_generator = tasker.run_yield()
        dependency_tasker_generators = [(dependency_tasker, self.run_tasker(executor, session_config, manager,
                                                                            dependency_tasker.tasker, dependency_tasker.dependency_taskers))
                                        for dependency_tasker in dependency_taskers]
        while True:
            for dependency_tasker, dependency_tasker_generator in tuple(dependency_tasker_generators):
                try:
                    _thread_local.current_tasker = dependency_tasker.tasker
                    dependency_tasker_generator.send(None)
                except StopIteration:
                    if dependency_tasker.reduce_config:
                        dependency_tasker.run_reduce(executor, session_config, manager, dependency_tasker.arguments, True)
                    dependency_tasker_generators.remove((dependency_tasker, dependency_tasker_generator))

            try:
                _thread_local.current_tasker = self.tasker
                tasker_generator.send(None)
            except StopIteration:
                break
            if executor.env_variables.get("@streaming", False):
                yield TaskerYieldNext()

    def terminate(self):
        if not self.tasker:
            return
        self.tasker.terminate()
        self.tasker = None

    def compile_distinct_config(self, aggregate):
        subquery_name = "__subquery_" + str(uuid.uuid1().int) + "_distinct"
        config = copy.deepcopy(self.config)
        config.pop("loader", None)
        config.pop("loader_arguments", None)
        config.update({
            "input": "&.--." + subquery_name + "::" + self.config["output"].split("::")[-1].split(" ")[0],
            "output": self.config["output"],
            "querys": [],
            "schema": {},
            "intercept": None,
            "orders": [],
            "pipelines": [],
            "options": {},
            "dependencys": [],
            "states": [],
        })

        is_multi_group_key = True
        if not aggregate["key"]:
            if len(aggregate["distinct_keys"]) == 1:
                group_column, is_multi_group_key = aggregate["distinct_keys"][0], False
            else:
                group_column = ["#make", aggregate["distinct_keys"], [":@aggregate_key", "$.*"]]
        elif isinstance(aggregate["key"], list) and aggregate["key"][0] == "#make":
            group_column = copy.deepcopy(aggregate["key"])
            group_column[1].extend(aggregate["distinct_keys"])
        else:
            group_column = ["#make", [copy.deepcopy(aggregate["key"])] + aggregate["distinct_keys"], [":@aggregate_key", "$.*"]]
        distinct_aggregate = copy.deepcopy(DEAULT_AGGREGATE)
        for key, column in tuple(self.config["schema"].items()):
            if key in aggregate["distinct_aggregates"]:
                config["schema"][key] = ["#make", {
                    "key": ["@aggregate_key", "$._aggregate_distinct_key_"] if is_multi_group_key else "$._aggregate_distinct_key_",
                    "value": "$." + key
                }, aggregate["schema"][key]["aggregate"]]
                self.config["schema"][key] = aggregate["schema"][key]["value"]
            elif key in aggregate["schema"]:
                self.config["schema"][key] = ["#make", {
                    "key": copy.deepcopy(group_column),
                    "value": aggregate["schema"][key]["value"]
                }, aggregate["schema"][key]["aggregate"]]
                distinct_aggregate["schema"][key] = copy.deepcopy(aggregate["schema"][key])
                distinct_aggregate["schema"][key]["final_value"] = None
                if is_multi_group_key:
                    config["schema"][key] = ["#aggregate", ["@aggregate_key", "$._aggregate_distinct_key_"],
                                             aggregate["schema"][key]["reduce"]]
                else:
                    config["schema"][key] = ["#aggregate", "$._aggregate_distinct_key_", aggregate["schema"][key]["reduce"]]
            else:
                config["schema"][key] = "$." + key

        if aggregate["key"]:
            self.config["schema"]["_aggregate_distinct_key_"] = aggregate["key"]
            if is_multi_group_key:
                aggregate["key"] = ["@aggregate_key", "$._aggregate_distinct_key_"]
            else:
                aggregate["key"] = "$._aggregate_distinct_key_"
        self.config["schema"]["_aggregate_distinct_aggregate_key_"] = ["#make", {
            "key": copy.deepcopy(group_column),
            "value": ["#const", 0]
        }, [":#aggregate", "$.key", "$$.value"]]
        distinct_aggregate["key"] = copy.deepcopy(group_column)
        distinct_aggregate["schema"]["_aggregate_distinct_aggregate_key_"] = {
            "key": group_column,
            "value": ["#const", 0],
            "calculate": ["#const", 0],
            "aggregate": [":#aggregate", "$.key", ["#const", 0]],
            "reduce": ["#const", 0],
            "final_value": None
        }
        self.config["aggregate"] = distinct_aggregate
        config["predicate"] = self.config.pop("predicate", None)
        self.config["output"] = "&.--." + subquery_name + "::" + self.config["output"].split("::")[-1].split(" ")[0] + " use I"
        self.config.pop("outputer", None)
        self.config.pop("outputer_arguments", None)
        return config

    def compile_reduce_config(self, aggregate):
        subquery_name = "__subquery_" + str(uuid.uuid1().int) + "_reduce"
        config = copy.deepcopy(self.config)
        config.pop("loader", None)
        config.pop("loader_arguments", None)
        config.update({
            "input": "&.--." + subquery_name + "::" + self.config["output"].split("::")[-1].split(" ")[0],
            "output": self.config["output"],
            "querys": [],
            "predicate": None,
            "schema": {},
            "orders": [],
            "pipelines": [],
            "options": {},
            "dependencys": [],
            "states": [],
        })

        is_multi_group_key = False
        if isinstance(aggregate["key"], list):
            if aggregate["key"] and aggregate["key"][0] == "@aggregate_key":
                is_multi_group_key = True
            elif len(aggregate["key"]) == 3 and aggregate["key"][0] == "#make" and aggregate["key"][2][0] == ":@aggregate_key":
                is_multi_group_key = True
        for key, column in self.config["schema"].items():
            if key in aggregate["schema"]:
                if is_multi_group_key:
                    config["schema"][key] = ["#aggregate", ["@aggregate_key", "$._aggregate_key_"], aggregate["schema"][key]["reduce"]]
                else:
                    config["schema"][key] = ["#aggregate", "$._aggregate_key_", aggregate["schema"][key]["reduce"]]
            else:
                config["schema"][key] = "$." + key
        if aggregate["key"]:
            config["schema"]["_aggregate_key_"] = "$._aggregate_key_"
            self.config["schema"]["_aggregate_key_"] = aggregate["key"]
        config["aggregate"] = aggregate
        config["intercept"] = self.config.pop("intercept", None)
        config["pipelines"] = self.config.pop("pipelines", [])
        self.config["output"] = "&.--." + subquery_name + "::" + self.config["output"].split("::")[-1].split(" ")[0] + " use I"
        self.config.pop("outputer", None)
        self.config.pop("outputer_arguments", None)
        self.reduce_config = config

    def run_reduce(self, executor, session_config, manager, arguments, final_reduce=False):
        config, arguments = copy.deepcopy(self.reduce_config), copy.deepcopy(arguments)
        aggregate = config.pop("aggregate", None)
        if final_reduce:
            config["schema"].pop("_aggregate_key_", None)
            for key, column in config["schema"].items():
                if key in aggregate["schema"] and aggregate["schema"][key]["final_value"]:
                    config["schema"][key] = aggregate["schema"][key]["final_value"]
                else:
                    config["schema"][key] = "$." + key
            for key, column in aggregate["window_schema"].items():
                if isinstance(column["aggregate"], dict):
                    for window_aggregate_key in column["aggregate"]:
                        config["schema"].pop(window_aggregate_key, None)
                    config["schema"][key] = column["final_value"]
            config["name"] = config["name"] + "#select@final"
        else:
            config["output"] = config["input"] + " use DI"
            config["name"] = config["name"] + "#select@reduce"
            config["intercept"] = None
            config["pipelines"] = []

        tasker = CoreTasker(config, manager)
        arguments["@primary_order"] = False
        arguments["@limit"] = 0
        self.compile_tasker(arguments, tasker)
        tasker_generator = self.run_tasker(executor, session_config, manager, tasker, [])
        while True:
            try:
                tasker_generator.send(None)
            except StopIteration:
                break
        if not final_reduce:
            self.tasker.status["store_count"] = tasker.status["store_count"]

    def compile_tasker(self, arguments, tasker):
        tasker.load()
        compile_arguments = dict(**self.ENV_ARGUMENTS)
        compile_arguments.update(arguments)
        if self.is_local_memory_database(tasker.config):
            if isinstance(tasker.config["input"], str) and "&.--." in tasker.config["input"] and not arguments.get("@loop", False):
                compile_arguments["@batch"] = 0
            if isinstance(tasker.config["output"], str) and "&.--." in tasker.config["output"]:
                compile_arguments["@insert_batch"] = 0

        tasker.compile(compile_arguments)
        if "@verbose" in compile_arguments and compile_arguments["@verbose"]:
            warp_database_logging(tasker)

        if hasattr(tasker.loader, "name"):
            if tasker.loader.name.startswith("--.__subquery_") or tasker.loader.name.startswith("--.__unionquery_"):
                self.temporary_memory_manager.use_collection(tasker.loader.name)

        if hasattr(tasker.outputer, "name"):
            if tasker.outputer.name.startswith("--.__subquery_") or tasker.outputer.name.startswith("--.__unionquery_"):
                self.temporary_memory_manager.add_collection(tasker.outputer.name)

        loader_or_outputers = [tasker.loader, tasker.outputer] + list(tasker.join_loaders.values())
        for loader_or_outputer in loader_or_outputers:
            if not hasattr(loader_or_outputer, "name"):
                continue
            info = loader_or_outputer.name.split(".")
            if len(info) <= 1:
                continue
            db, table_name = info[0], info[1]
            if not table_name.startswith("#{env_variable__@") or table_name[-1] != "}":
                continue
            if not loader_or_outputer.name.startswith(db + "."):
                continue

            def env_variable_setter(ctasker, ctable_name, obj, prefix, key):
                def _(executor, session_config, manager):
                    table_name_value = str(executor.env_variables.get_value(key))
                    if ctable_name in ctasker.input:
                        ctasker.input = ctasker.input.replace(ctable_name, table_name_value)
                    if ctable_name in ctasker.output:
                        ctasker.output = ctasker.output.replace(ctable_name, table_name_value)
                    obj.name = prefix + table_name_value
                return _
            self.updaters.append(env_variable_setter(tasker, table_name, loader_or_outputer, db + ".", table_name[16:-1]))
        tasker.name = "[%s]%s" % (self.tasker_index, tasker.name)
        setattr(tasker, "running_id", self.tasker_index)
        return compile_arguments

    def run_tasker(self, executor, session_config, manager, tasker, dependency_taskers):
        try:
            tasker_generator = self.run_yield(executor, session_config, manager, tasker, dependency_taskers)
            while True:
                try:
                    value = tasker_generator.send(None)
                    if isinstance(value, TaskerYieldNext):
                        yield value
                except StopIteration:
                    break
        except SystemError as e:
            self.check_free_temporary_memory_collection(executor, tasker)
            tasker.close(False, "signal terminaled")
            get_logger().error("signal exited")
            raise e
        except KeyboardInterrupt as e:
            self.check_free_temporary_memory_collection(executor, tasker)
            tasker.close(False, "user terminaled")
            get_logger().error("Crtl+C exited")
            raise e
        except Exception as e:
            self.check_free_temporary_memory_collection(executor, tasker)
            tasker.close(False, "Error: " + repr(e), traceback.format_exc())
            get_logger().error("tasker %s error: %s\n%s", tasker.name, e, traceback.format_exc())
            raise e
        else:
            self.check_free_temporary_memory_collection(executor, tasker)
            tasker.close()

    def execute_updater(self, executor, session_config, manager):
        for updater in self.updaters:
            updater(executor, session_config, manager)

        for dependency_tasker in self.dependency_taskers:
            dependency_tasker.execute_updater(executor, session_config, manager)

    def get_temporary_memory_collections(self):
        names = self.temporary_memory_manager.get_names()
        for dependency_tasker in self.dependency_taskers:
            names.extend(dependency_tasker.get_temporary_memory_collections())
        return list(set(names))

    def check_free_temporary_memory_collection(self, executor, tasker):
        try:
            if not hasattr(tasker.loader, "name"):
                return
            if not tasker.loader.name.startswith("--.__subquery_") and not tasker.loader.name.startswith("--.__unionquery_"):
                return
            if hasattr(tasker.outputer, "name") and tasker.loader.name == tasker.outputer.name:
                return
            name = tasker.loader.name
            if self.temporary_memory_manager.unuse_collection(name):
                database = tasker.databases.instance("--")
                if executor.runners and database.is_streaming(name):
                    return
                delete = database.delete(name, ["id"])
                delete.commit()
                self.temporary_memory_manager.free_collection(name)
        except:
            pass

    def is_local_memory_database(self, config, name="--"):
        try:
            database_config = dict(**[database for database in config["databases"]
                                      if database["name"] == name][0])
            return database_config["driver"] == "memory"
        except (TypeError, KeyError, IndexError):
            return False



# -*- coding: utf-8 -*-
# 2023/4/24
# create by: snower

from .example import ExampleTestCase


class LoopExampleTestCase(ExampleTestCase):
    example_name = "loop"

    def test_loop(self):
        self.execute("loop.sql")

        self.assert_data(15, [{'n': 1}, {'n': 2}, {'n': 3}, {'n': 4}], "data error")


# -*- coding: utf-8 -*-
# 2023/4/25
# create by: snower

from .example import ExampleTestCase


class YieldDataExampleTestCase(ExampleTestCase):
    example_name = "yield_data"

    def test_type_annotation(self):
        self.execute("yield_data.sql")

        self.assert_data(5, [{'a': 'a', 'b': 1}, {'a': 'b', 'b': 2}, {'a': 'c', 'b': 3}], "data error")

        self.assert_data(11, [{'a': '青菜', 'b': '青菜'}, {'a': '白菜', 'b': '白菜'}, {'a': '青菜', 'b': '青菜'}], "data error")

        self.assert_data(13, [{'a': 1}, {'a': 2}, {'a': 3}], "data error")

        self.assert_data(17, [{'RANGE_COUNT(3)': 0, 'GENERATE_CUSTOMIZE$RANGE_COUNT(4)': 0},
                              {'RANGE_COUNT(3)': 1, 'GENERATE_CUSTOMIZE$RANGE_COUNT(4)': 1},
                              {'RANGE_COUNT(3)': 2, 'GENERATE_CUSTOMIZE$RANGE_COUNT(4)': 2},
                              {'RANGE_COUNT(3)': None, 'GENERATE_CUSTOMIZE$RANGE_COUNT(4)': 3}], "data error")
